#!/usr/bin/python

""" Reference implementation of the Python XML Templating Language.
"""

__all__    = ['processFile', 'Transformer']

import os, sys, types, operator, string, urllib, urlparse
from pxtl.constants import *
from pxtl.exceptions import *
from pxtl.serialisers import *
from pxtl.utilities import *

def processFile(
  path, writer= None, globals= None, debug= False, headers= False, dom= None
):
  """ Read, deserialise and transform a template and any subtemplates it
      imports, from a file in the filesystem. Optional arguments:

      writer  - stream-like object to send serialised output (default stdout)
      globals - dictionary to use as template global scope (default {})
      debug   - output a debugging page if there's an error (default False)
      headers - output HTTP response headers (for CGI scripts, default False)
      dom     - DOM L3 Core/XML/Load implementation to use (default pxdom)
  """
  if writer is None:
    writer= sys.stdout
  if dom is None:
    from pxtl import pxdom
    dom= pxdom.getDOMImplementation('')
  parser= dom.createLSParser(dom.MODE_SYNCHRONOUS, None)
  parser.config.setParameter('entities', True)
  parser.config.setParameter('cdata-sections', True)
  t= Transformer()
  uri= 'file:'+urllib.pathname2url(path)
  try:
    try:
      doc= parser.parseURI(uri)
    except NonErrors:
      raise
    except Exception, e:
      raise PXTLParseError(e, uri)
    doc= t.transformDocument(doc, globals)
    if not serialisers.has_key(t.method):
      raise PXTLMethodValueError(t.method)
  except PXTLError, e:
    if not debug:
      raise
    if headers:
      writer.write('Status: 500 Server error\r\n')
    processFile(ERROR_PAGE, writer, {'e': e}, False, headers, dom)
    return
  encoding= None
  if UnicodeType is not None:
    try:
      encoding= doc.inputEncoding
    except AttributeError:
      pass
    if encoding is None:
      try:
        encoding= doc.xmlEncoding
      except AttributeError:
        pass
    if encoding is None:
      encoding= 'utf-8'
  if headers:
    mimetype= t.mimetype
    if mimetype is None:
      if t.method in ('html', 'xhtml'):
        mimetype= 'text/html'
      elif t.method=='xml':
        mimetype= 'text/xml'
      else:
        mimetype= 'text/plain'
    if encoding is not None and (
      mimetype[:5]=='text/' or mimetype[-4:]=='+xml'
    ):
      writer.write('Content-Type: %s; charset=%s\r\n\r\n'%(mimetype,encoding))
    else:
      writer.write('Content-Type: %s\r\n\r\n' % mimetype)
  ser= serialisers[t.method](writer, encoding)
  ser.writeDocument(doc, t.literals)


class Transformer:
  """ Effect PXTL transformations on a DOM Level 2 Core-compliant DOM tree.
  """
  def __init__(self):
    self.pxtl= PXTLObject()

  def transformDocument(self, document, globals= None):
    """ Main public interface. PXTL-transform an XML Document node, executing
        Python code with the optional dictionary 'globals' as global scope.
        Returns the transformed Document node, which may or may not be the
        original node, depending on DOM. Places the template's desired output
        method in the Transformer's 'method' property and a list of
        pre-serialised text nodes (from px_mark PI etc) in the property
        'literals', which the caller may read if interested.
    """
    root= document.documentElement
    dom= document.implementation

    # Try to make a Level 3 LSParser object so we can load new documents. If
    # this is not doable, px:import elements will not work.
    #
    self.parser= None
    try:
      self.parser= dom.createLSParser(dom.MODE_SYNCHRONOUS, None)
      self.parser.config.setParameter('entities', True)
      self.parser.config.setParameter('cdata-sections', True)
    except AttributeError:
      pass

    # Make a new variable-environment-holding scoper object. If a global scope
    # was supplied, store it in the scoper. If the nested_scope feature is on
    # (either because it is always on in this version of Python, or the user
    # asked for it in the future attribute), ask the scoper to use nesteds.
    #
    if globals is None:
      globals= {}
    globals['pxtl']= self.pxtl
    locals= None
    features= checkFeatures(root.getAttributeNodeNS(PXNS, 'future'))
    if haveNestedScopes(features):
      nesteds= []
    else:
      nesteds= None
    self.scoper= Scoper(globals, locals, nesteds, features)

    # Work out what doctype and output method the output document will be,
    # from the PXTL doctype attribute on the root.
    #
    self.mimetype, self.method, publicId, systemId= (None,)*4
    doctype= root.getAttributeNodeNS(PXNS, 'doctype')
    if doctype is not None:
      doctuple= self.scoper.pxeval(doctype.value, doctype)
      if type(doctuple)!=types.TupleType:
        raise PXTLDocumentTypeError(doctype, doctuple)
      self.mimetype, self.method, publicId, systemId= (doctuple+(None,)*4)[:4]

    # If not specified, use no doctype, and guess between xml and xhtml output
    # methods.
    #
    if self.method is None:
      ns= root.namespaceURI
      if ns==HTNS or (ns is None and root.localName=='html'):
        self.method= 'xhtml'
      else:
        self.method= 'xml'

    # Write to doctype; this may cause a new document to be created.
    #
    if doctype is not None:
      doctype= dom.createDocumentType(root.tagName, publicId, systemId)
    document= setDoctype(document, doctype)

    # Run transformations on the document and return the document shorn of any
    # reference to now-unused PXTL namespaces.
    #
    self.literals= []
    self.elementContext= ConditionalContext()
    self.attributeContext= ConditionalContext()
    self.whitespace= True
    self.transformChildren(document)
    removePXTLNamespaces(document)
    return document


  # Transformation of DOM Nodes.
  #
  def transformChildren(self, node, rslice= None):
    """ Apply PXTL transformations to all child nodes, or to the last 'rslice'
        child nodes if supplied.
    """
    # Despatch each transformable child to relevant function.
    #
    children= list(node.childNodes)
    if rslice is not None:
      children[:len(children)-rslice]= []
    elementContext= self.elementContext
    attributeContext= self.attributeContext
    self.elementContext= ConditionalContext()
    for child in children:
      if child.nodeType==node.ELEMENT_NODE:
        self.attributeContext= attributeContext.copy()
        whitespace= self.whitespace
        self.transformElement(child)
        self.whitespace= whitespace
      elif child.nodeType==node.PROCESSING_INSTRUCTION_NODE:
        self.transformPI(child)
    self.elementContext= elementContext
    self.attributeContext= attributeContext


  def transformElement(self, node):
    """ Perform all PXTL transformations on an element.
    """
    # Ignore/remove doctype attribute (only used by processDocument function
    # in root element of root template), and future attribute (only used in
    # root element of any template).
    #
    attr= node.getAttributeNodeNS(PXNS, 'doctype')
    if attr is not None:
      node.removeAttributeNode(attr)
    attr= node.getAttributeNodeNS(PXNS, 'future')
    if attr is not None:
      node.removeAttributeNode(attr)

    # If there is a PXTL space attribute on the element, change whitespace
    # processing mode.
    #
    attr= node.getAttributeNodeNS(PXNS, 'space')
    if attr is not None:
      whitespace= self.scoper.pxeval(attr.value, attr)
      if whitespace is not None:
        self.whitespace= [True, False][not whitespace]
      node.removeAttributeNode(attr)

    # If we are in no-whitespace mode, remove all whitespace in child element
    # content not surrounded by text. This assumes the document has been read
    # in a normal form, ie. there is only one Text node between non-Text
    # nodes.
    #
    if not self.whitespace:
      for child in list(node.childNodes):
        if child.nodeType==child.TEXT_NODE:
          data=child.data
          while len(data)>0 and data[0] in ' \n\t':
            data= data[1:]
          while len(data)>0 and data[-1] in ' \n\t':
            data= data[:-1]
          if data=='':
            node.removeChild(child)
          else:
            child.data= data

    # Change tag name if tagname attribute present.
    #
    attr= node.getAttributeNodeNS(PXNS, 'tagname')
    if attr is not None:
      qname= self.scoper.pxeval(attr.value, attr)
      if not isinstance(qname, StringTypes) or qname=='':
        raise PXTLTagnameAttrError(attr, qname)

      # Work out what namespace the element with a possibly-changed prefix
      # would be in.
      #
      if string.count(qname, ':')==1 and ':' not in qname[1]+qname[-1]:
        ns= node.lookupNamespaceURI(string.split(qname, ':')[0])
      else:
        ns= node.lookupNamespaceURI(None)
      if ns in [PXNS, NSNS]:
        raise PXTLTagnamespaceAttrError(attr, ns)

      # Make a new element with the desired tagname, then move all attributes
      # and children from the old node to the new one. (tagname is readonly.)
      #
      newel= node.ownerDocument.createElementNS(ns, qname)
      for attr in attrList(node.attributes):
        if not (attr.namespaceURI==PXNS and attr.localName=='tagname'):
          newel.setAttributeNodeNS(attr.cloneNode())
      for child in list(node.childNodes):
        newel.appendChild(child)
      node.parentNode.replaceChild(newel, node)

    # If the tagname is px:none, element is never successful. Otherwise look
    # for conditional attributes to determine success. Remove unsuccessful
    # elements.
    #
    if node.namespaceURI==PXNS and node.localName=='none':
      success= False
    else:
      success= None
      attrs= attrList(node.attributes)
      for attr in attrs:
        cond= attr.localName
        if attr.namespaceURI==PXNS and cond in CONDITIONALS:
          if success is not None:
            raise PXTLMultipleConditionalError(attr.ownerElement)
          testing= self.attributeContext.getTesting(cond)
          if testing is None:
            raise PXTLConditionalOrderError(
              attr, self.attributeContext.old_cond
            )
          testvalue= None
          if testing:
            testvalue= True
            if cond!='else':
              testvalue= self.scoper.pxeval(attr.value, attr)
          success= self.attributeContext.getSuccess(cond, testvalue)
          node.removeAttributeNode(attr)
    if success==False:
      self.transformChildren(node)
      removeReparent(node)
      return

    # Check all attributes for pseudo-PIs.
    #
    for attr in attrList(node.attributes):
      if string.find(attr.value, '{?')!=-1:
        parts= parsePseudoPIs(attr)

        # If there are any px_if pseudo-PIs that evaluate False, throw the
        # attribute away before evaluating anything else.
        #
        success= True
        for part in parts:
          if type(part)==types.TupleType and part[0]=='px_if':
            if not self.scoper.pxeval(part[1], attr):
              success= False
              break
        if not success:
          node.removeAttributeNode(attr)
          continue

        # Replace all other pseudo-PIs with their values.
        #
        for i in range(len(parts)):
          if type(parts[i])==types.TupleType:
            (target, data)= parts[i]
            if target=='px_if':
              parts[i]= ''
              continue
            elif SHORTCUTS.has_key(target):
              target= SHORTCUTS[target]
            elif target[:3]=='px_':
              target= target[3:]
            else:
              raise PXTLPseudoPITargetPXError(attr, target)
            if target=='note':
              parts[i]= ''
            elif target in CODINGS:
              d= self.scoper.pxeval(data, attr)
              if not isinstance(d, StringTypes):
                d= str(d)
              parts[i]= encodeText(d, target)
            else:
              raise PXTLPseudoPITargetNameError(attr, target)

        # Write the finished attribute back to the node.
        #
        attr.value= string.join(parts, '')

    # If there is a PXTL attr attribute, a dictionary, evaluate it and send
    # all its values to new/replacement attributes on the element.
    #
    attr= node.getAttributeNodeNS(PXNS, 'attr')
    if attr is not None:
      dict= self.scoper.pxeval(attr.value, attr)
      if type(dict)!=types.DictionaryType:
        raise PXTLAttrTypeError(attr, dict)
      for (qname, value) in dict.items():
        if not isinstance(qname, StringTypes):
          raise PXTLAttrNameTypeError(attr, qname)
        if not isinstance(value, StringTypes):
          value= str(value)
        if string.count(qname, ':')==1 and ':' not in qname[1]+qname[-1]:
          ns= node.lookupNamespaceURI(string.split(qname, ':')[0])
        else:
          ns= node.lookupNamespaceURI(None)
        newattr= node.ownerDocument.createAttributeNS(ns, qname)
        newattr.value= value
        node.setAttributeNodeNS(newattr)
      node.removeAttributeNode(attr)

    # Handle PXTL elements.
    #
    if node.namespaceURI==PXNS and node.localName in CONDITIONALS:
      self.conditionalElement(node)
    else:
      self.elementContext.reset()
      if node.namespaceURI==PXNS:
        if node.localName=='for':
          self.forElement(node)
        elif node.localName=='while':
          self.whileElement(node)
        elif node.localName=='def':
          self.defElement(node)
        elif node.localName=='call':
          self.callElement(node)
        elif node.localName=='import':
          self.importElement(node)
        else:
          raise PXTLTagNameError(node)
      else:
        self.transformChildren(node)


  def transformPI(self, node):
    """ Perform PXTL transformations on a Processing Instruction.
    """
    # Only interested in PIs with target name px_..., or shortcuts for them.
    # Complain about px_... targets we do not recognise.
    #
    if SHORTCUTS.has_key(node.target):
      target= SHORTCUTS[node.target]
    elif node.target[:3]=='px_':
      target= node.target[3:]
    else:
      return
    if target not in TARGETS:
      raise PXTLPITargetNameError(node)

    # Just delete comment PIs.
    #
    if target=='note':
      node.parentNode.removeChild(node)

    # Execute PIs that produce output. If not mark-up, encode the output.
    # Replace the node with any output produced by the PI.
    #
    elif target in CODINGS+['mark']:
      d= self.scoper.pxeval(node.data, node)
      if not isinstance(d, StringTypes):
        d= str(d)
      if d=='':
        node.parentNode.removeChild(node)
      else:
        if target!='mark':
          d= encodeText(d, target)
        output= node.ownerDocument.createTextNode(d)
        if target=='mark':
          self.literals.append(output)
        node.parentNode.replaceChild(output, node)

    # Execute the code PI. Generates no markup directly, but may call write()
    # on the pxtl object to output literal and/or encoded output. Check the
    # pxtl object after executing the code and grabs anything that has been
    # output, to replace the original node.
    #
    elif target=='code':
      self.pxtl._begin()
      if node.data[:2]==':\n':
        self.scoper.pxblock(node.data[2:], node)
      else:
        self.scoper.pxexec(string.strip(node.data), node)
      texts= self.pxtl._read()
      for (literal, text) in texts:
        output= node.ownerDocument.createTextNode(text)
        if literal:
          self.literals.append(output)
        node.parentNode.insertBefore(output, node)
      node.parentNode.removeChild(node)
    else:
      raise PXTLPITargetNameError(node)


  # Specific PXTL elements
  #
  def conditionalElement(self, node):
    """ Simple conditional element. Remove it from the document if test False.
    """
    test= None
    cond= node.localName
    if cond!='else':
      test= node.getAttributeNodeNS(None, 'test')
      if test is None:
        raise PXTLAttributeMissingError(node, 'test')

    testing= self.elementContext.getTesting(cond)
    if testing is None:
      raise PXTLConditionalOrderError(node, self.elementContext.old_cond)

    testvalue= None
    if testing:
      testvalue= True
      if cond!='else':
        testvalue= operator.truth(self.scoper.pxeval(test.value, test))

    success= self.elementContext.getSuccess(cond, testvalue)
    if success:
      self.transformChildren(node)
      removeReparent(node)
    else:
      node.parentNode.removeChild(node)


  def forElement(self, node):
    """ Include and transform child nodes once for each item in a list,
        assigning loop variable each time around.
    """
    # Move loop body into a holding fragment.
    #
    body= node.ownerDocument.createDocumentFragment()
    for child in list(node.childNodes):
      body.appendChild(child)
    children= list(body.childNodes)

    # Get iterable object from range or items attribute.
    #
    rattr= node.getAttributeNodeNS(None, 'range')
    iattr= node.getAttributeNodeNS(None, 'in')
    if (rattr is None) == (iattr is None):
      raise PXTLForItemsError(node)
    if iattr is not None:
      items= self.scoper.pxeval(iattr.value, iattr)
    else:
      r= self.scoper.pxeval(rattr.value, rattr)
      if type(r) in (types.IntType, types.LongType):
        r= (r,)
      if type(r)!=types.TupleType:
        raise PXTLForRangeError(rattr, r)
      items= apply(xrange, r)

    # Iterate over items, optionally assigning iteration item and index vars.
    #
    item= node.getAttributeNodeNS(None, 'item')
    index= node.getAttributeNodeNS(None, 'index')
    success= len(items)>0
    for i in xrange(len(items)):
      if item is not None:
        self.scoper.pxassign(item.value, items[i], item)
      if index is not None:
        self.scoper.pxassign(index.value, i, item)

      # Each time round the loop, make a copy of the body inside the loop and
      # transform it.
      #
      for child in children:
        node.appendChild(child.cloneNode(True))
      self.transformChildren(node, rslice= len(children))

    # Finished, remove the template's for element leaving the iteration
    # results at this level.
    #
    self.elementContext.getSuccess('for', success)
    removeReparent(node)


  def whileElement(self, node):
    """ Repeatly include and transform child elements.
    """
    # Get test attribute and optional min, index.
    #
    test= node.getAttributeNodeNS(None, 'test')
    if test is None:
      raise PXTLAttributeMissingError(node, 'test')
    attr= node.getAttributeNodeNS(None, 'min')
    if attr is None:
      min= 0
    else:
      min= self.scoper.pxeval(attr.value, attr)
      if not isinstance(min, IntTypes):
        raise PXTLWhileMinError(attr, min)
    index= node.getAttributeNodeNS(None, 'index')

    # Move loop body into a holding fragment.
    #
    body= node.ownerDocument.createDocumentFragment()
    for child in list(node.childNodes):
      body.appendChild(child)
    children= list(body.childNodes)

    # Continue until test condition evaluates False. Do not test the condition
    # for the first min times around the loop. Assign loop index if index
    # attribute used.
    #
    i= 0
    success= False
    while True:
      if i>=min:
        if not self.scoper.pxeval(test.value, test):
          break
      success= True
      if index is not None:
        self.scoper.pxassign(index.value, i, index)
      i= i+1

      # Make a copy of the loop body in the loop element and apply PXTL
      # transformations.
      #
      for child in children:
        node.appendChild(child.cloneNode(True))
      self.transformChildren(node, rslice= len(children))

    # Finished, remove loop element.
    #
    self.elementContext.getSuccess('while', success)
    removeReparent(node)


  def defElement(self, node):
    """ Define a subtemplate that can be called like a function.
    """
    if not node.hasAttributeNS(None, 'fn'):
      raise PXTLAttributeMissingError(node, 'fn')
    fn= node.getAttributeNS(None, 'fn')
    args= None
    if node.hasAttributeNS(None, 'args'):
      args= node.getAttributeNS(None, 'args')

    # Move subtemplate definition into a holding fragment.
    #
    body= node.ownerDocument.createDocumentFragment()
    for child in list(node.childNodes):
      body.appendChild(child)

    # Make the subtemplate object which will simulate function calls. Store
    # the currently active namespaces in it so they are restored when the
    # subtemplate is called in a different part of the document.
    #
    t= Subtemplate(body, args, getNamespaces(node), self.scoper, node)
    t.documentURI= node.ownerDocument.documentURI

    # Put the subtemplate in scope and remove its definition node.
    #
    self.scoper.pxassign(fn, t, node)
    node.parentNode.removeChild(node)


  def callElement(self, node):
    """ Call a previously-defined subtemplate.
    """
    # Get the subtemplate object referred to.
    #
    fn= node.getAttributeNodeNS(None, 'fn')
    if fn is None:
      raise PXTLAttributeMissingError(node, 'fn')
    fn= self.scoper.pxeval(fn.value, fn)
    if not isinstance(fn, Subtemplate):
      raise PXTLCallTypeError(node, fn)

    # Restore any namespaces that were defined on the def element but are
    # different now.
    #
    nss= getNamespaces(node)
    for (prefix, uri) in fn.nss.items():
      if not nss.has_key(prefix) or nss[prefix]!=uri:
        qname= 'xmlns'
        if prefix is not None:
          qname= 'xmlns:'+prefix
        node.setAttributeNS(NSNS, qname, uri)

    # Assign the argument list to the appropriate variables in a new nested
    # scope.
    #
    args= None
    if node.hasAttributeNS(None, 'args'):
      args= node.getAttributeNS(None, 'args')
    parentScoper= self.scoper
    parentDocumentURI= node.ownerDocument.documentURI
    self.scoper= fn.scoperWithArgs(args, parentScoper, node)
    node.ownerDocument.documentURI= fn.documentURI

    # Copy the holding fragment back into the document at the point of the
    # call element. Apply PXTL transformations using the new nested scope and
    # remove the call element.
    #
    for child in fn.body.childNodes:
      node.appendChild(child.cloneNode(True))
    self.transformChildren(node)
    node.ownerDocument.documentURI= parentDocumentURI
    self.scoper= parentScoper
    removeReparent(node)


  def importElement(self, node):
    """ Import element handler. Replace import element with either text from
        the imported resource, or the DOM of a PXTL template, depending on the
        type of the imported resource.
    """
    if self.parser is None:
      raise PXTLNoParserError(node)
    src= node.getAttributeNodeNS(None, 'src')
    if src is None:
      raise PXTLAttributeMissingError(node, 'src')
    src= self.scoper.pxeval(src.value, src)
    if node.ownerDocument.documentURI is not None:
      src= urlparse.urljoin(node.ownerDocument.documentURI, src)

    # If type attribute specified, decide how to include based on that,
    # otherwise try to work out what type the resource is by asking urllib.
    # Will almost always use the filename extension of local name.
    #
    t= node.getAttributeNodeNS(None, 'type')
    if t is not None:
      t= self.scoper.pxeval(t.value, t)
    else:
      try:
        stream= urllib.urlopen(src)
      except IOError:
        raise PXTLImportSrcError(node, src)
      t= stream.info().gettype()
      stream.close()

    # For non-PXTL includes, read the resource and include it as text. If it's
    # an XML document already, include it as literal pre-serialised content.
    #
    if t in TYPES_TEXT+TYPES_XML:
      try:
        stream= urllib.urlopen(src)
      except IOError:
        raise PXTLImportSrcError(node, src)
      content= stream.read()
      stream.close()
      text= node.ownerDocument.createTextNode(content)
      node.parentNode.replaceChild(text, node)
      if t in TYPES_XML:
        self.literals.append(text)

    # For PXTL includes, get the resource as a DOM Document and copy its
    # transformed content to the current document.
    #
    else:
      try:
        idoc= self.parser.parseURI(src)
      except IOError:
        raise PXTLImportSrcError(node, src)
      except Exception, e:
        raise PXTLParseError(e, src)
      iroot= idoc.documentElement

      # Get the new global scope to be used in this file from the globals
      # attribute, or use a new blank dictionary if none is supplied.
      #
      globalsAttr= node.getAttributeNodeNS(None, 'globals')
      if globalsAttr is None:
        globalsDict= {}
      else:
        globalsDict= self.scoper.pxeval(globalsAttr.value, globalsAttr)
        if type(globalsDict)!=types.DictType:
          raise PXTLImportGlobalsError(globalsAttr, globalsDict)
      globalsDict['pxtl']= self.pxtl

      # Get features of imported document and make a new Scoper to hold them
      # and the global scope.
      #
      features= checkFeatures(iroot.getAttributeNodeNS(PXNS, 'future'))
      if haveNestedScopes(features):
        nesteds= []
      else:
        nesteds= None
      nestedScoper= Scoper(globalsDict, None, nesteds, features)

      # If 'as' attribute passed, write what will be the global scope of the
      # imported file to the given l-value in the importing scope.
      #
      as= node.getAttributeNodeNS(None, 'as')
      if as is not None:
        self.scoper.pxassign(as.value, DictAsInstance(globalsDict), as)

      # Import each root-level node of the document into a new fragment. Apply
      # PXTL transformations and put the results into the importing document.
      #
      frag= node.ownerDocument.createDocumentFragment()
      for child in list(idoc.childNodes):
        if child.nodeType!=child.DOCUMENT_TYPE_NODE:
          frag.appendChild(node.ownerDocument.importNode(child, True))
      _scoper= self.scoper
      _whitespace= self.whitespace
      _elementContext= self.elementContext
      _attributeContext= self.elementContext
      _documentURI= node.ownerDocument.documentURI
      self.scoper= nestedScoper
      self.whitespace= True
      self.elementContext= ConditionalContext()
      self.attributeContext= ConditionalContext()
      node.ownerDocument.documentURI= src

      self.transformChildren(frag)

      self.scoper= _scoper
      self.whitespace= _whitespace
      self.elementContext= _elementContext
      self.attributeContext= _attributeContext
      node.ownerDocument.documentURI= _documentURI

      node.parentNode.replaceChild(frag, node)


# Reference implementation utility objects
#

class PXTLObject(AbstractPXTLObject):
  def write(self, text, coding='text'):
    """ Can only be called in a px_code PI, otherwise pxtlobject is not
        writeable. Adds possibly-encoded text to a list of text nodes to
        create. For 'mark' encoding the output will end up in a literal
        node.
    """
    if not self.writeable:
      raise Exception('pxtl.write can only be called in code PIs')
    if not isinstance(text, StringTypes):
      text= str(text)
    literal= coding=='mark'
    if not literal:
      text= encodeText(text, coding)
    if len(self.texts)>0 and self.texts[-1][0]==literal:
      self.texts[-1]= (literal, self.texts[-1][1]+text)
    else:
      self.texts.append((literal, text))

  def __init__(self):
    self.writeable= False

  def _begin(self):
    self.writeable= True
    self.texts= []

  def _read(self):
    self.writeable= False
    return self.texts


class Subtemplate:
  """ Class representing a subtemplate function. Used to interpret parameter
      lists.
  """
  def __init__(self, body, args, nss, scoper, context):
    self.body= body
    self.nss= nss
    self.scoper= scoper
    self.args= args
    if args is not None:
      code= 'def __pxtl_assign(%s): return locals()' % args
      scoper.pxexec(code, context)
      locals= scoper.getLocals()
      self.readargs= locals['__pxtl_assign']
      del locals['__pxtl_assign']

  def scoperWithArgs(self, args, scoper, context):
    if (args is None) != (self.args is None):
      raise PXTLCallArgsError(context, args)
    if args is None:
      l= {}
    else:
      locals= scoper.getLocals()
      locals['__pxtl_assign']= self.readargs
      code= '__pxtl_assign= __pxtl_assign(%s)' % args
      scoper.pxexec(code, context)
      l= locals['__pxtl_assign']
      del locals['__pxtl_assign']
    return self.scoper.makeNestedScoper(l)


class Scoper:
  """ Represents a code block's variable environment. Includes a local scope,
      global scope, and in Python 2.2 and later (or 2.1 with a suitable future
      declaration) a set of nested scopes. Also holds, but does not do
      anything with, a list of future-features.
  """
  def __init__(self, globals, locals, nesteds, features):
    """ Construct a new scoper with optional local scope and a (possibly
        empty) list of nested scopes. If nesteds is 'None' instead of a list,
        nested scopes are disabled.
    """
    self.globals= globals
    self.locals= locals
    self.nesteds= nesteds
    self.features= features

  def makeNestedScoper(self, locals):
    """ Return a scoper that will emulate the variable environment of a
        function nested inside the current environment. The new local scope is
        supplied as a parameter. Global scope is the same as the current
        scoper's. If nested scopes are enabled the old local scope is added to
        list of nested scopes in the new scoper.
    """
    if self.nesteds is None:
      return Scoper(self.globals, locals, None, self.features)
    if self.locals is None:
      return Scoper(self.globals, locals, [], self.features)
    nesteds= self.nesteds+[self.locals]
    return Scoper(self.globals, locals, nesteds, self.features)

  def getLocals(self):
    """ Return the locals dictionary, or globals if there is none.
    """
    if self.locals is not None:
      return self.locals
    else:
      return self.globals

  def getGlobals(self):
    """ Return a globals dictionary with any nested scopes combined into it.
    """
    if self.nesteds is None or len(self.nesteds)==0:
      return self.globals
    globals= self.globals.copy()
    for nested in self.nesteds:
      globals.update(nested)
    return globals

  # Python wrappers. Run statements, expressions and assignments in an
  # environment described by the current scoper object. Handle and re-raise
  # exceptions as PXTL runtime errors.
  #
  def pxexec(self, code, context):
    """ Wrapper for 'exec'.
    """
    code= string.lstrip(code)
    if code=='':
      return
    if len(self.features)>0:
      features= string.join(self.features, ', ')
      code= 'from __future__ import %s\n%s' % (features, code)
    try:
      if convertExecs:
        code= str(code)
      exec code in self.getGlobals(), self.getLocals()
    except NonErrors+(PXTLError,):
      raise
    except:
      raise PXTLRuntimeError(context, sys.exc_info(), code)


  def pxeval(self, expr, context):
    """ Wrapper for 'eval'. (Can't use eval itself as might need to include
        __future__ declaration.)
    """
    globals= self.getGlobals()
    locals= self.getLocals()
    code= '__pxtl_assign= %s\n' % string.lstrip(expr)
    if len(self.features)>0:
      features= string.join(self.features, ', ')
      code= 'from __future__ import %s\n%s' % (features, code)
    try:
      if convertExecs:
        code= str(code)
      exec code in globals, self.getLocals()
    except NonErrors+(PXTLError,):
      raise
    except:
      raise PXTLRuntimeError(context, sys.exc_info(), code)
    value= locals['__pxtl_assign']
    del locals['__pxtl_assign']
    return value


  def pxassign(self, var, value, context):
    """ Assign something to an l-value given in a string expression. (Can't
        write directly to scope as l-value might not be a simple variable.)
    """
    globals= self.getGlobals()
    globals['__pxtl_assign']= value
    code= '%s= __pxtl_assign\n' % string.lstrip(var)
    if len(self.features)>0:
      features= string.join(self.features, ', ')
      code= 'from __future__ import %s\n%s' % (features, code)
    try:
      if convertExecs:
        code= str(code)
      exec code in globals, self.getLocals()
    except NonErrors+(PXTLError,):
      raise
    except:
      raise PXTLRuntimeError(context, sys.exc_info(), code)
    del globals['__pxtl_assign']


  def pxblock(self, code, context):
    """ Execute multiple lines of code indented with an arbitrary amount of
        whitespace.
    """
    # Find first non-empty, non-comment line and see if it is indented.
    #
    indented= False
    for line in string.split(code, '\n'):
      stripped= string.lstrip(line)
      if stripped!='' and stripped[0]!='#':
        indented= len(stripped)!=len(line)
        break

    # If the block is indented, put it in a dummy construct that requires
    # indentation.
    #
    if indented:
      code= 'if 1:\n%s\nelse: pass\n' % code
    if len(self.features)>0:
      features= string.join(self.features, ', ')
      code= 'from __future__ import %s\n%s' % (features, code)
    try:
      if convertExecs:
        code= str(code)
      exec code in self.getGlobals(), self.getLocals()
    except NonErrors+(PXTLError,):
      raise
    except:
      raise PXTLRuntimeError(context, sys.exc_info(), code)


def removePXTLNamespaces(node):
  """ Once all PXTL elements and attributes have been processed away, remove
      namespace declarations that refer to PXTL.
  """
  if node.attributes is not None:
    for attr in attrList(node.attributes):
      if attr.namespaceURI==NSNS:
        if attr.value==PXNS:
          node.removeAttributeNode(attr)
  for child in list(node.childNodes):
    if child.nodeType==child.ELEMENT_NODE:
      removePXTLNamespaces(child)
